@@ -39,17 +39,17 @@ static int	version_f(int argc, char **argv);
 static void     version_help(void);
 
 static const cmdinfo_t	sb_cmd =
-	{ "sb", NULL, sb_f, 0, 1, 1, "[agno]",
-	  "set current address to sb header", sb_help };
+	{ "sb", NULL, sb_f, 0, 1, 1, N_("[agno]"),
+	  N_("set current address to sb header"), sb_help };
 static const cmdinfo_t	uuid_cmd =
-	{ "uuid", NULL, uuid_f, 0, 1, 1, "[uuid]",
-	  "write/print FS uuid", uuid_help };
+	{ "uuid", NULL, uuid_f, 0, 1, 1, N_("[uuid]"),
+	  N_("write/print FS uuid"), uuid_help };
 static const cmdinfo_t	label_cmd =
-	{ "label", NULL, label_f, 0, 1, 1, "[label]",
-	  "write/print FS label", label_help };
+	{ "label", NULL, label_f, 0, 1, 1, N_("[label]"),
+	  N_("write/print FS label"), label_help };
 static const cmdinfo_t	version_cmd =
-	{ "version", NULL, version_f, 0, -1, 1, "[feature | [vnum fnum]]",
-	  "set feature bit(s) in the sb version field", version_help };
+	{ "version", NULL, version_f, 0, -1, 1, N_("[feature | [vnum fnum]]"),
+	  N_("set feature bit(s) in the sb version field"), version_help };
 
 void
 sb_init(void)
@@ -108,7 +108,19 @@ const field_t	sb_flds[] = {
 	{ "logsectsize", FLDT_UINT16D, OI(OFF(logsectsize)), C1, 0, TYP_NONE },
 	{ "logsunit", FLDT_UINT32D, OI(OFF(logsunit)), C1, 0, TYP_NONE },
 	{ "features2", FLDT_UINT32X, OI(OFF(features2)), C1, 0, TYP_NONE },
-	{ "bad_features2", FLDT_UINT32X, OI(OFF(bad_features2)), C1, 0, TYP_NONE },
+	{ "bad_features2", FLDT_UINT32X, OI(OFF(bad_features2)),
+		C1, 0, TYP_NONE },
+	{ "features_compat", FLDT_UINT32X, OI(OFF(features_compat)),
+		C1, 0, TYP_NONE },
+	{ "features_ro_compat", FLDT_UINT32X, OI(OFF(features_ro_compat)),
+		C1, 0, TYP_NONE },
+	{ "features_incompat", FLDT_UINT32X, OI(OFF(features_incompat)),
+		C1, 0, TYP_NONE },
+	{ "features_log_incompat", FLDT_UINT32X, OI(OFF(features_log_incompat)),
+		C1, 0, TYP_NONE },
+	{ "crc", FLDT_CRC, OI(OFF(crc)), C1, 0, TYP_NONE },
+	{ "pquotino", FLDT_INO, OI(OFF(pquotino)), C1, 0, TYP_INODE },
+	{ "lsn", FLDT_UINT64X, OI(OFF(lsn)), C1, 0, TYP_NONE },
 	{ NULL }
 };
 
@@ -120,7 +132,7 @@ const field_t	sb_hfld[] = {
 static void
 sb_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " set allocation group superblock\n"
 "\n"
@@ -134,7 +146,7 @@ sb_help(void)
 " remaining allocation groups only serve as backup for filesystem recovery.\n"
 " The icount/ifree/fdblocks/frextents are only updated in superblock 0.\n"
 "\n"
-);
+));
 }
 
 static int
@@ -148,7 +160,7 @@ sb_f(
 	if (argc > 1) {
 		agno = (xfs_agnumber_t)strtoul(argv[1], &p, 0);
 		if (*p != '\0' || agno >= mp->m_sb.sb_agcount) {
-			dbprintf("bad allocation group number %s\n", argv[1]);
+			dbprintf(_("bad allocation group number %s\n"), argv[1]);
 			return 0;
 		}
 		cur_agno = agno;
@@ -180,78 +192,72 @@ get_sb(xfs_agnumber_t agno, xfs_sb_t *sb)
 		XFS_FSS_TO_BB(mp, 1), DB_RING_IGN, NULL);
 
 	if (!iocur_top->data) {
-		dbprintf("can't read superblock for AG %u\n", agno);
+		dbprintf(_("can't read superblock for AG %u\n"), agno);
 		pop_cur();
 		return 0;
 	}
 
-	libxfs_xlate_sb(iocur_top->data, sb, 1, XFS_SB_ALL_BITS);
+	libxfs_sb_from_disk(sb, iocur_top->data);
 
 	if (sb->sb_magicnum != XFS_SB_MAGIC) {
-		dbprintf("bad sb magic # %#x in AG %u\n",
+		dbprintf(_("bad sb magic # %#x in AG %u\n"),
 			sb->sb_magicnum, agno);
 		return 0;
 	}
-	if (!XFS_SB_GOOD_VERSION(sb)) {
-		dbprintf("bad sb version # %#x in AG %u\n",
+	if (!xfs_sb_good_version(sb)) {
+		dbprintf(_("bad sb version # %#x in AG %u\n"),
 			sb->sb_versionnum, agno);
 		return 0;
 	}
 	if (agno == 0 && sb->sb_inprogress != 0) {
-		dbprintf("mkfs not completed successfully\n");
+		dbprintf(_("mkfs not completed successfully\n"));
 		return 0;
 	}
 	return 1;
 }
 
 /* workaround craziness in the xlog routines */
-int xlog_recover_do_trans(xlog_t *log, xlog_recover_t *t, int p) { return 0; }
+int xlog_recover_do_trans(struct xlog *log, xlog_recover_t *t, int p)
+{
+	return 0;
+}
 
 int
 sb_logcheck(void)
 {
-	xlog_t		log;
-	xfs_daddr_t	head_blk, tail_blk;
+	int		dirty;
 
 	if (mp->m_sb.sb_logstart) {
 		if (x.logdev && x.logdev != x.ddev) {
-			dbprintf("aborting - external log specified for FS "
-				 "with an internal log\n");
+			dbprintf(_("aborting - external log specified for FS "
+				 "with an internal log\n"));
 			return 0;
 		}
 	} else {
 		if (!x.logdev || (x.logdev == x.ddev)) {
-			dbprintf("aborting - no external log specified for FS "
-				 "with an external log\n");
+			dbprintf(_("aborting - no external log specified for FS "
+				 "with an external log\n"));
 			return 0;
 		}
 	}
 
-	memset(&log, 0, sizeof(log));
-	if (!x.logdev)
-		x.logdev = x.ddev;
-	x.logBBsize = XFS_FSB_TO_BB(mp, mp->m_sb.sb_logblocks);
-	x.logBBstart = XFS_FSB_TO_DADDR(mp, mp->m_sb.sb_logstart);
-	log.l_dev = (mp->m_sb.sb_logstart == 0) ? x.logdev : x.ddev;
-	log.l_logsize = BBTOB(log.l_logBBsize);
-	log.l_logBBsize = x.logBBsize;
-	log.l_logBBstart = x.logBBstart;
-	log.l_mp = mp;
+	libxfs_buftarg_init(mp, x.ddev, x.logdev, x.rtdev);
 
-	if (xlog_find_tail(&log, &head_blk, &tail_blk, 0)) {
-		dbprintf("ERROR: cannot find log head/tail, run xfs_repair\n");
+	dirty = xlog_is_dirty(mp, &x, 0);
+	if (dirty == -1) {
+		dbprintf(_("ERROR: cannot find log head/tail, run xfs_repair\n"));
 		return 0;
-	}
-	if (head_blk != tail_blk) {
-		dbprintf(
+	} else if (dirty == 1) {
+		dbprintf(_(
 "ERROR: The filesystem has valuable metadata changes in a log which needs to\n"
 "be replayed.  Mount the filesystem to replay the log, and unmount it before\n"
 "re-running %s.  If you are unable to mount the filesystem, then use\n"
 "the xfs_repair -L option to destroy the log and attempt a repair.\n"
 "Note that destroying the log may cause corruption -- please attempt a mount\n"
-"of the filesystem before doing this.\n", progname);
+"of the filesystem before doing this.\n"), progname);
 		return 0;
 	}
+	/* Log is clean */
 	return 1;
 }
 
@@ -261,16 +267,15 @@ sb_logzero(uuid_t *uuidp)
 	if (!sb_logcheck())
 		return 0;
 
-	dbprintf("Clearing log and setting UUID\n");
+	dbprintf(_("Clearing log and setting UUID\n"));
 
-	if (libxfs_log_clear(
-			(mp->m_sb.sb_logstart == 0) ? x.logdev : x.ddev,
+	if (libxfs_log_clear(mp->m_logdev_targp,
 			XFS_FSB_TO_DADDR(mp, mp->m_sb.sb_logstart),
 			(xfs_extlen_t)XFS_FSB_TO_BB(mp, mp->m_sb.sb_logblocks),
 			uuidp,
-			XFS_SB_VERSION_HASLOGV2(&mp->m_sb) ? 2 : 1,
+			xfs_sb_version_haslogv2(&mp->m_sb) ? 2 : 1,
 			mp->m_sb.sb_logsunit, XLOG_FMT)) {
-		dbprintf("ERROR: cannot clear the log\n");
+		dbprintf(_("ERROR: cannot clear the log\n"));
 		return 0;
 	}
 	return 1;
@@ -280,7 +285,7 @@ sb_logzero(uuid_t *uuidp)
 static void
 uuid_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " write/print FS uuid\n"
 "\n"
@@ -298,7 +303,7 @@ uuid_help(void)
 "As a side effect of writing the UUID, the log is cleared (which is fine\n"
 "on a CLEANLY unmounted FS).\n"
 "\n"
-);
+));
 }
 
 static uuid_t *
@@ -317,7 +322,7 @@ do_uuid(xfs_agnumber_t agno, uuid_t *uuid)
 	}
 	/* set uuid */
 	memcpy(&tsb.sb_uuid, uuid, sizeof(uuid_t));
-	libxfs_xlate_sb(iocur_top->data, &tsb, -1, XFS_SB_UUID);
+	libxfs_sb_to_disk(iocur_top->data, &tsb, XFS_SB_UUID);
 	write_cur();
 	return uuid;
 }
@@ -333,14 +338,26 @@ uuid_f(
 	uuid_t		*uup = NULL;
 
 	if (argc != 1 && argc != 2) {
-		dbprintf("invalid parameters\n");
+		dbprintf(_("invalid parameters\n"));
 		return 0;
 	}
 
 	if (argc == 2) {	/* WRITE UUID */
 
 		if ((x.isreadonly & LIBXFS_ISREADONLY) || !expert_mode) {
-			dbprintf("%s: not in expert mode, writing disabled\n",
+			dbprintf(_("%s: not in expert mode, writing disabled\n"),
+				progname);
+			return 0;
+		}
+
+		/*
+		 * For now, changing the UUID of V5 superblock filesystems is
+		 * not supported; we do not have the infrastructure to fix all
+		 * other metadata when a new superblock UUID is generated.
+		 */
+		if (xfs_sb_version_hascrc(&mp->m_sb) &&
+		    strcasecmp(argv[1], "rewrite")) {
+			dbprintf(_("%s: only 'rewrite' supported on V5 fs\n"),
 				progname);
 			return 0;
 		}
@@ -352,32 +369,32 @@ uuid_f(
 		} else if (!strcasecmp(argv[1], "rewrite")) {
 			uup = do_uuid(0, NULL);
 			if (!uup) {
-				dbprintf("failed to read UUID from AG 0\n");
+				dbprintf(_("failed to read UUID from AG 0\n"));
 				return 0;
 			}
 			memcpy(&uu, uup, sizeof(uuid_t));
 			platform_uuid_unparse(&uu, bp);
-			dbprintf("old UUID = %s\n", bp);
+			dbprintf(_("old UUID = %s\n"), bp);
 		} else {
 			if (platform_uuid_parse(argv[1], &uu)) {
-				dbprintf("invalid UUID\n");
+				dbprintf(_("invalid UUID\n"));
 				return 0;
 			}
 		}
 
-		/* clear the log (setting uuid) if its not dirty */
+		/* clear the log (setting uuid) if it's not dirty */
 		if (!sb_logzero(&uu))
 			return 0;
 
-		dbprintf("writing all SBs\n");
+		dbprintf(_("writing all SBs\n"));
 		for (agno = 0; agno < mp->m_sb.sb_agcount; agno++)
 			if (!do_uuid(agno, &uu)) {
-				dbprintf("failed to set UUID in AG %d\n", agno);
+				dbprintf(_("failed to set UUID in AG %d\n"), agno);
 				break;
 			}
 
 		platform_uuid_unparse(&uu, bp);
-		dbprintf("new UUID = %s\n", bp);
+		dbprintf(_("new UUID = %s\n"), bp);
 		return 0;
 
 	} else {	/* READ+CHECK UUID */
@@ -385,14 +402,14 @@ uuid_f(
 		for (agno = 0; agno < mp->m_sb.sb_agcount; agno++) {
 			uup = do_uuid(agno, NULL);
 			if (!uup) {
-				dbprintf("failed to read UUID from AG %d\n",
+				dbprintf(_("failed to read UUID from AG %d\n"),
 					agno);
 				return 0;
 			}
 			if (agno) {
 				if (memcmp(&uu, uup, sizeof(uuid_t))) {
-					dbprintf("warning: UUID in AG %d "
-						 "differs to the primary SB\n",
+					dbprintf(_("warning: UUID in AG %d "
+						 "differs to the primary SB\n"),
 						agno);
 					break;
 				}
@@ -402,15 +419,15 @@ uuid_f(
 		}
 		if (mp->m_sb.sb_logstart) {
 			if (x.logdev && x.logdev != x.ddev)
-				dbprintf("warning - external log specified "
-					 "for FS with an internal log\n");
+				dbprintf(_("warning - external log specified "
+					 "for FS with an internal log\n"));
 		} else if (!x.logdev || (x.logdev == x.ddev)) {
-			dbprintf("warning - no external log specified "
-				 "for FS with an external log\n");
+			dbprintf(_("warning - no external log specified "
+				 "for FS with an external log\n"));
 		}
 
 		platform_uuid_unparse(&uu, bp);
-		dbprintf("UUID = %s\n", bp);
+		dbprintf(_("UUID = %s\n"), bp);
 	}
 
 	return 0;
@@ -420,7 +437,7 @@ uuid_f(
 static void
 label_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " write/print FS label\n"
 "\n"
@@ -435,7 +452,7 @@ label_help(void)
 "specified value.  The maximum length of a label is 12 characters - use of a\n"
 "longer label will result in truncation and a warning will be issued.\n"
 "\n"
-);
+));
 }
 
 static char *
@@ -458,7 +475,7 @@ do_label(xfs_agnumber_t agno, char *label)
 	/* set label */
 	if ((len = strlen(label)) > sizeof(tsb.sb_fname)) {
 		if (agno == 0)
-			dbprintf("%s: truncating label length from %d to %d\n",
+			dbprintf(_("%s: truncating label length from %d to %d\n"),
 				progname, (int)len, (int)sizeof(tsb.sb_fname));
 		len = sizeof(tsb.sb_fname);
 	}
@@ -470,7 +487,7 @@ do_label(xfs_agnumber_t agno, char *label)
 	memset(&tsb.sb_fname, 0, sizeof(tsb.sb_fname));
 	memcpy(&tsb.sb_fname, label, len);
 	memcpy(&lbl[0], &tsb.sb_fname, sizeof(tsb.sb_fname));
-	libxfs_xlate_sb(iocur_top->data, &tsb, -1, XFS_SB_FNAME);
+	libxfs_sb_to_disk(iocur_top->data, &tsb, XFS_SB_FNAME);
 	write_cur();
 	return &lbl[0];
 }
@@ -485,40 +502,40 @@ label_f(
 	xfs_agnumber_t	ag;
 
 	if (argc != 1 && argc != 2) {
-		dbprintf("invalid parameters\n");
+		dbprintf(_("invalid parameters\n"));
 		return 0;
 	}
 
 	if (argc == 2) {	/* WRITE LABEL */
 
 		if ((x.isreadonly & LIBXFS_ISREADONLY) || !expert_mode) {
-			dbprintf("%s: not in expert mode, writing disabled\n",
+			dbprintf(_("%s: not in expert mode, writing disabled\n"),
 				progname);
 			return 0;
 		}
 
-		dbprintf("writing all SBs\n");
+		dbprintf(_("writing all SBs\n"));
 		for (ag = 0; ag < mp->m_sb.sb_agcount; ag++)
 			if ((p = do_label(ag, argv[1])) == NULL) {
-				dbprintf("failed to set label in AG %d\n", ag);
+				dbprintf(_("failed to set label in AG %d\n"), ag);
 				break;
 			}
-		dbprintf("new label = \"%s\"\n", p);
+		dbprintf(_("new label = \"%s\"\n"), p);
 
 	} else {	/* READ LABEL */
 
 		for (ag = 0; ag < mp->m_sb.sb_agcount; ag++) {
 			p = do_label(ag, NULL);
 			if (!p) {
-				dbprintf("failed to read label in AG %d\n", ag);
+				dbprintf(_("failed to read label in AG %d\n"), ag);
 				return 0;
 			}
 			if (!ag)
 				memcpy(&sb.sb_fname, p, sizeof(sb.sb_fname));
 			else if (memcmp(&sb.sb_fname, p, sizeof(sb.sb_fname)))
-				dbprintf("warning: AG %d label differs\n", ag);
+				dbprintf(_("warning: AG %d label differs\n"), ag);
 		}
-		dbprintf("label = \"%s\"\n", p);
+		dbprintf(_("label = \"%s\"\n"), p);
 	}
 	return 0;
 }
@@ -527,7 +544,7 @@ label_f(
 static void
 version_help(void)
 {
-	dbprintf(
+	dbprintf(_(
 "\n"
 " set/print feature bits in sb version\n"
 "\n"
@@ -540,11 +557,11 @@ version_help(void)
 " 'version log2'     - enable v2 log format\n"
 "\n"
 "The version function prints currently enabled features for a filesystem\n"
-"according to its the version field of the primary superblock.\n"
+"according to the version field of its primary superblock.\n"
 "It can also be used to enable selected features, such as support for\n"
-"unwritten extents.  The upated version is written into all AGs.\n"
+"unwritten extents.  The updated version is written into all AGs.\n"
 "\n"
-);
+));
 }
 
 static int
@@ -556,16 +573,23 @@ do_version(xfs_agnumber_t agno, __uint16_t version, __uint32_t features)
 	if (!get_sb(agno, &tsb))
 		return 0;
 
+	if (xfs_sb_has_mismatched_features2(&tsb)) {
+		dbprintf(_("Superblock has mismatched features2 fields, "
+			   "skipping modification\n"));
+		return 0;
+	}
+
 	if ((version & XFS_SB_VERSION_LOGV2BIT) &&
-	    !XFS_SB_VERSION_HASLOGV2(&tsb)) {
+					!xfs_sb_version_haslogv2(&tsb)) {
 		tsb.sb_logsunit = 1;
 		fields |= (1LL << XFS_SBS_LOGSUNIT);
 	}
 
 	tsb.sb_versionnum = version;
 	tsb.sb_features2 = features;
-	fields |= XFS_SB_VERSIONNUM | XFS_SB_FEATURES2;
-	libxfs_xlate_sb(iocur_top->data, &tsb, -1, fields);
+	tsb.sb_bad_features2 = features;
+	fields |= XFS_SB_VERSIONNUM | XFS_SB_FEATURES2 | XFS_SB_BAD_FEATURES2;
+	libxfs_sb_to_disk(iocur_top->data, &tsb, fields);
 	write_cur();
 	return 1;
 }
@@ -584,38 +608,54 @@ version_string(
 		strcpy(s, "V3");
 	else if (XFS_SB_VERSION_NUM(sbp) == XFS_SB_VERSION_4)
 		strcpy(s, "V4");
+	else if (XFS_SB_VERSION_NUM(sbp) == XFS_SB_VERSION_5)
+		strcpy(s, "V5");
 
-	if (XFS_SB_VERSION_HASATTR(sbp))
+	if (xfs_sb_version_hasattr(sbp))
 		strcat(s, ",ATTR");
-	if (XFS_SB_VERSION_HASNLINK(sbp))
+	if (xfs_sb_version_hasnlink(sbp))
 		strcat(s, ",NLINK");
-	if (XFS_SB_VERSION_HASQUOTA(sbp))
+	if (xfs_sb_version_hasquota(sbp))
 		strcat(s, ",QUOTA");
-	if (XFS_SB_VERSION_HASALIGN(sbp))
+	if (xfs_sb_version_hasalign(sbp))
 		strcat(s, ",ALIGN");
-	if (XFS_SB_VERSION_HASDALIGN(sbp))
+	if (xfs_sb_version_hasdalign(sbp))
 		strcat(s, ",DALIGN");
-	if (XFS_SB_VERSION_HASSHARED(sbp))
+	if (xfs_sb_version_hasshared(sbp))
 		strcat(s, ",SHARED");
-	if (XFS_SB_VERSION_HASDIRV2(sbp))
+	if (xfs_sb_version_hasdirv2(sbp))
 		strcat(s, ",DIRV2");
-	if (XFS_SB_VERSION_HASLOGV2(sbp))
+	if (xfs_sb_version_haslogv2(sbp))
 		strcat(s, ",LOGV2");
-	if (XFS_SB_VERSION_HASEXTFLGBIT(sbp))
+	if (xfs_sb_version_hasextflgbit(sbp))
 		strcat(s, ",EXTFLG");
-	if (XFS_SB_VERSION_HASSECTOR(sbp))
+	if (xfs_sb_version_hassector(sbp))
 		strcat(s, ",SECTOR");
 	if (xfs_sb_version_hasasciici(sbp))
 		strcat(s, ",ASCII_CI");
-	if (XFS_SB_VERSION_HASMOREBITS(sbp))
+	if (xfs_sb_version_hasmorebits(sbp))
 		strcat(s, ",MOREBITS");
-	if (XFS_SB_VERSION_HASATTR2(sbp))
+	if (xfs_sb_version_hasattr2(sbp))
 		strcat(s, ",ATTR2");
-	if (XFS_SB_VERSION_LAZYSBCOUNT(sbp))
+	if (xfs_sb_version_haslazysbcount(sbp))
 		strcat(s, ",LAZYSBCOUNT");
+	if (xfs_sb_version_hasprojid32bit(sbp))
+		strcat(s, ",PROJID32BIT");
+	if (xfs_sb_version_hascrc(sbp))
+		strcat(s, ",CRC");
+	if (xfs_sb_version_hasftype(sbp))
+		strcat(s, ",FTYPE");
+	if (xfs_sb_version_hasfinobt(sbp))
+		strcat(s, ",FINOBT");
 	return s;
 }
 
+/*
+ * XXX: this only supports reading and writing to version 4 superblock fields.
+ * V5 superblocks always define certain V4 feature bits - they are blocked from
+ * being changed if a V5 sb is detected, but otherwise v5 superblock features
+ * are not handled here.
+ */
 static int
 version_f(
 	int		argc,
@@ -628,7 +668,7 @@ version_f(
 	if (argc == 2) {	/* WRITE VERSION */
 
 		if ((x.isreadonly & LIBXFS_ISREADONLY) || !expert_mode) {
-			dbprintf("%s: not in expert mode, writing disabled\n",
+			dbprintf(_("%s: not in expert mode, writing disabled\n"),
 				progname);
 			return 0;
 		}
@@ -646,13 +686,17 @@ version_f(
 				version = 0x0034 | XFS_SB_VERSION_EXTFLGBIT;
 				break;
 			case XFS_SB_VERSION_4:
-				if (XFS_SB_VERSION_HASEXTFLGBIT(&mp->m_sb))
-					dbprintf("unwritten extents flag"
-						 " is already enabled\n");
+				if (xfs_sb_version_hasextflgbit(&mp->m_sb))
+					dbprintf(
+		_("unwritten extents flag is already enabled\n"));
 				else
 					version = mp->m_sb.sb_versionnum |
 						  XFS_SB_VERSION_EXTFLGBIT;
 				break;
+			case XFS_SB_VERSION_5:
+				dbprintf(
+		_("unwritten extents always enabled for v5 superblocks.\n"));
+				break;
 			}
 		} else if (!strcasecmp(argv[1], "log2")) {
 			switch (XFS_SB_VERSION_NUM(&mp->m_sb)) {
@@ -666,41 +710,55 @@ version_f(
 				version = 0x0034 | XFS_SB_VERSION_LOGV2BIT;
 				break;
 			case XFS_SB_VERSION_4:
-				if (XFS_SB_VERSION_HASLOGV2(&mp->m_sb))
-					dbprintf("version 2 log format"
-						 " is already in use\n");
+				if (xfs_sb_version_haslogv2(&mp->m_sb))
+					dbprintf(
+		_("version 2 log format is already in use\n"));
 				else
 					version = mp->m_sb.sb_versionnum |
 						  XFS_SB_VERSION_LOGV2BIT;
 				break;
+			case XFS_SB_VERSION_5:
+				dbprintf(
+		_("Version 2 logs always enabled for v5 superblocks.\n"));
+				break;
 			}
+		} else if (XFS_SB_VERSION_NUM(&mp->m_sb) == XFS_SB_VERSION_5) {
+			dbprintf(
+		_("%s: Cannot change %s on v5 superblocks.\n"),
+				progname, argv[1]);
+			return 0;
 		} else if (!strcasecmp(argv[1], "attr1")) {
-			if (XFS_SB_VERSION_HASATTR2(&mp->m_sb)) {
+
+			if (xfs_sb_version_hasattr2(&mp->m_sb)) {
 				if (!(mp->m_sb.sb_features2 &=
 						~XFS_SB_VERSION2_ATTR2BIT))
 					mp->m_sb.sb_versionnum &=
 						~XFS_SB_VERSION_MOREBITSBIT;
 			}
-			XFS_SB_VERSION_ADDATTR(&mp->m_sb);
+			xfs_sb_version_addattr(&mp->m_sb);
 			version = mp->m_sb.sb_versionnum;
 			features = mp->m_sb.sb_features2;
 		} else if (!strcasecmp(argv[1], "attr2")) {
-			XFS_SB_VERSION_ADDATTR(&mp->m_sb);
-			XFS_SB_VERSION_ADDATTR2(&mp->m_sb);
+			xfs_sb_version_addattr(&mp->m_sb);
+			xfs_sb_version_addattr2(&mp->m_sb);
+			version = mp->m_sb.sb_versionnum;
+			features = mp->m_sb.sb_features2;
+		} else if (!strcasecmp(argv[1], "projid32bit")) {
+			xfs_sb_version_addprojid32bit(&mp->m_sb);
 			version = mp->m_sb.sb_versionnum;
 			features = mp->m_sb.sb_features2;
 		} else {
-			dbprintf("%s: invalid version change command \"%s\"\n",
+			dbprintf(_("%s: invalid version change command \"%s\"\n"),
 				progname, argv[1]);
 			return 0;
 		}
 
 		if (version) {
-			dbprintf("writing all SBs\n");
+			dbprintf(_("writing all SBs\n"));
 			for (ag = 0; ag < mp->m_sb.sb_agcount; ag++)
 				if (!do_version(ag, version, features)) {
-					dbprintf("failed to set versionnum "
-						 "in AG %d\n", ag);
+					dbprintf(_("failed to set versionnum "
+						 "in AG %d\n"), ag);
 					break;
 				}
 			mp->m_sb.sb_versionnum = version;
@@ -717,7 +775,7 @@ version_f(
 		mp->m_sb.sb_features2 = strtoul(argv[2], &sp, 0);
 	}
 
-	dbprintf("versionnum [0x%x+0x%x] = %s\n", mp->m_sb.sb_versionnum,
+	dbprintf(_("versionnum [0x%x+0x%x] = %s\n"), mp->m_sb.sb_versionnum,
 			mp->m_sb.sb_features2, version_string(&mp->m_sb));
 
 	if (argc == 3) {	/* now reset... */
